/*
 * ALSA driver for Panasonic UniPhier series.
 * 
 * Copyright (c) 2013 Panasonic corporation.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/irqshdrv.h>

#include "mn2ws-pcm.h"

MODULE_AUTHOR("Katsuhiro Suzuki <suzuki.katsuhiro002@jp.panasonic.com>");
MODULE_DESCRIPTION("Panasonic UniPhier MN2WS0230 PCM Driver(AIO direct)");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE(
	"{{Panasonic, MN2WS0230}, "
	"{Panasonic, MN2WS0230x}}");


#define MN2WS_PCM_AIO_DEV_NAME       "mn2ws0230-pcm-aio"
#define MN2WS_PCM_AIO_DEV_COUNT      3

#define MN2WS_RES_PCM            0
#define MN2WS_RES_AIO2           1

#define MN2WS_PCM_REG_START      (0x5e000000)
#define MN2WS_PCM_REG_SIZE       (0x00012000)
#define MN2WS_AIO2_REG_START     (0x56000000)
#define MN2WS_AIO2_REG_SIZE      (0x00050000)

#define MN2WS_PCM_HW_BUF_SIZE    (PAGE_SIZE * 8)

//FIXME: tentative
#define MN2WS_PCM_AIO_DUMMY_CH   (3)


//APCMOUT
#define MN2WS_APCMOUT_MODE(n)             (0x5e011730 + 0x30 * (n))
#define MN2WS_APCMOUT_SEMAPH(n)           (0x5e011734 + 0x30 * (n))
#define MN2WS_APCMOUT_DEVSEL_CONF(n)      (0x5e011738 + 0x30 * (n))
#define MN2WS_APCMOUT_DEVSEL_SEND(n)      (0x5e01173c + 0x30 * (n))
#define MN2WS_APCMOUT_SAMP(n)             (0x5e011740 + 0x30 * (n))

#define MN2WS_APCMOUT_SEND_STADDR(n)      (0x5e011b30 + 0x30 * (n))
#define MN2WS_APCMOUT_SEND_ENDADDR(n)     (0x5e011b34 + 0x30 * (n))
#define MN2WS_APCMOUT_SEND_RDSRC(n)       (0x5e011b38 + 0x30 * (n))
#define MN2WS_APCMOUT_SEND_WRSRC(n)       (0x5e011b3c + 0x30 * (n))
#define MN2WS_APCMOUT_SEND_XOCTL(n)       (0x5e011b40 + 0x30 * (n))
#define MN2WS_APCMOUT_SEND_XOFS(n)        (0x5e011b44 + 0x30 * (n))

#define MN2WS_APCMOUT_SEND_HDMIPCMOCTL    (0x5e011ba8)
#define MN2WS_APCMOUT_SEND_HDMIIECOCTL    (0x5e011bac)
#define MN2WS_APCMOUT_SEND_HDMIFMTSEL     (0x5e011bb0)

#define MN2WS_APCMOUT_SEND_DUMMY0FS       (0x5e011bd0)

//Other
#define MN2WS_SYSCTRCMDA                  (0x5e00c09c)
#define MN2WS_ADECOUTCTL                  (0x5e010c00)
#define MN2WS_INTSTATAVIO6                (0x5e008658)


//AOUT(PCMOUTN) AOUTPORT_SW:0x56042000
#define MN2WS_A2SWOPortMXCtr1(n)               (0x56042000 + 0x400 * (n))
#define MN2WS_A2SWOPortMXCtr2(n)               (0x56042004 + 0x400 * (n))
#define MN2WS_A2SWOPortMXCtr3(n)               (0x56042008 + 0x400 * (n))
#define MN2WS_A2SWOPortMXACLKSel0Ex(n)         (0x56042030 + 0x400 * (n))
#define MN2WS_A2SWOPortMXPath(n)               (0x56042040 + 0x400 * (n))
#define MN2WS_A2SWOPortMXSYNC(n)               (0x56042044 + 0x400 * (n))
#define MN2WS_A2SWOPortMXCntCtr(n)             (0x56042060 + 0x400 * (n))
#define MN2WS_A2SWOPortMXCounter(n)            (0x56042064 + 0x400 * (n))
#define MN2WS_A2SWOPortMXIR(n)                 (0x56042070 + 0x400 * (n))
#define MN2WS_A2SWOPortMXIM0(n)                (0x56042080 + 0x400 * (n))
#define MN2WS_A2SWOPortMXIM1(n)                (0x56042084 + 0x400 * (n))
#define MN2WS_A2SWOPortMXIM2(n)                (0x56042088 + 0x400 * (n))
#define MN2WS_A2SWOPortMXIM3(n)                (0x5604208c + 0x400 * (n))
#define MN2WS_A2SWOPortMXID0(n)                (0x56042090 + 0x400 * (n))
#define MN2WS_A2SWOPortMXID1(n)                (0x56042094 + 0x400 * (n))
#define MN2WS_A2SWOPortMXID2(n)                (0x56042098 + 0x400 * (n))
#define MN2WS_A2SWOPortMXID3(n)                (0x5604209c + 0x400 * (n))
#define MN2WS_A2SWOPortMXEXNOE(n)              (0x560420f0 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0VolPara1(n)         (0x56042100 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1VolPara1(n)         (0x56042120 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2VolPara1(n)         (0x56042140 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3VolPara1(n)         (0x56042160 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4VolPara1(n)         (0x56042180 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0VolPara2(n)         (0x56042104 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1VolPara2(n)         (0x56042124 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2VolPara2(n)         (0x56042144 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3VolPara2(n)         (0x56042164 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4VolPara2(n)         (0x56042184 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0VolGainStatus(n)    (0x56042108 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1VolGainStatus(n)    (0x56042128 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2VolGainStatus(n)    (0x56042148 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3VolGainStatus(n)    (0x56042168 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4VolGainStatus(n)    (0x56042188 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0VolChPara(n)        (0x5604210c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1VolChPara(n)        (0x5604212c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2VolChPara(n)        (0x5604214c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3VolChPara(n)        (0x5604216c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4VolChPara(n)        (0x5604218c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0VolCtr(n)           (0x56042110 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1VolCtr(n)           (0x56042130 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2VolCtr(n)           (0x56042150 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3VolCtr(n)           (0x56042170 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4VolCtr(n)           (0x56042190 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0SlotCtr(n)          (0x56042114 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1SlotCtr(n)          (0x56042134 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2SlotCtr(n)          (0x56042154 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3SlotCtr(n)          (0x56042174 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4SlotCtr(n)          (0x56042194 + 0x400 * (n))
#define MN2WS_A2SWOPortMXT0RstCtr(n)           (0x5604211c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT1RstCtr(n)           (0x5604213c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT2RstCtr(n)           (0x5604215c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT3RstCtr(n)           (0x5604217c + 0x400 * (n))
#define MN2WS_A2SWOPortMXT4RstCtr(n)           (0x5604219c + 0x400 * (n))


//A2D(subsystem)
#define MN2WS_CDA2D_STRT0                 (0x56010000)
#define MN2WS_CDA2D_PERFCNFG              (0x56010008)
#define MN2WS_CDA2D_STAT0                 (0x56010020)
#define MN2WS_CDA2D_CHMXCTRL1(n)          (0x56011000 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXCTRL2(n)          (0x56011004 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXSTAT(n)           (0x56011010 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXIR(n)             (0x56011014 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXIE(n)             (0x56011018 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXID(n)             (0x5601101c + 0x40 * (n))
#define MN2WS_CDA2D_CHMXSRCAMODE(n)       (0x56011020 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXDSTAMODE(n)       (0x56011024 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXSRCSTRTADRS(n)    (0x56011028 + 0x40 * (n))
#define MN2WS_CDA2D_CHMXDSTSTRTADRS(n)    (0x5601102c + 0x40 * (n))
#define MN2WS_CDA2D_CHMXSIZE(n)           (0x56011030 + 0x40 * (n))

//A2D(ring buffer)
#define MN2WS_CDA2D_RBFLUSH0              (0x56010040)
#define MN2WS_CDA2D_PARTRESET0            (0x56010050)
#define MN2WS_CDA2D_PARTRESET1            (0x56010054)
#define MN2WS_CDA2D_RBMXBGNADRS(n)        (0x56012000 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXENDADRS(n)        (0x56012004 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXBTH(n)            (0x56012008 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXRTH(n)            (0x5601200c + 0x40 * (n))
#define MN2WS_CDA2D_RBMXSTAT(n)           (0x56012010 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXIR(n)             (0x56012014 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXIE(n)             (0x56012018 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXID(n)             (0x5601201c + 0x40 * (n))
#define MN2WS_CDA2D_RBMXRDPTR(n)          (0x56012020 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXWRPTR(n)          (0x56012028 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXCNFG(n)           (0x56012030 + 0x40 * (n))
#define MN2WS_CDA2D_RBMXPTRCTRL(n)        (0x56012034 + 0x40 * (n))


static int mn2ws0230_pcm_aio_hwdep_force_aout(struct mn2ws_pcm_dev *dev);

static int mn2ws0230_pcm_aio_play_setup(struct mn2ws_pcm_dev *dev);
static int mn2ws0230_pcm_aio_play_mute(struct mn2ws_pcm_dev *dev);
static int mn2ws0230_pcm_aio_play_start(struct mn2ws_pcm_dev *dev);
static int mn2ws0230_pcm_aio_play_stop(struct mn2ws_pcm_dev *dev);
static unsigned long mn2ws0230_pcm_aio_play_get_hwptr(struct mn2ws_pcm_dev *dev);
static void mn2ws0230_pcm_aio_play_set_devptr(struct mn2ws_pcm_dev *dev, unsigned long ptr);
static int mn2ws0230_pcm_aio_init_module(void);
static void mn2ws0230_pcm_aio_exit_module(void);


static struct resource mn2ws_resources[] = {
	[MN2WS_RES_PCM] = {
		.start = MN2WS_PCM_REG_START, 
		.end   = MN2WS_PCM_REG_START + MN2WS_PCM_REG_SIZE, 
		.flags = IORESOURCE_MEM, 
	}, 
	[MN2WS_RES_AIO2] = {
		.start = MN2WS_AIO2_REG_START, 
		.end   = MN2WS_AIO2_REG_START + MN2WS_AIO2_REG_SIZE, 
		.flags = IORESOURCE_MEM, 
	}, 
};

struct mn2ws_pcm_desc mn2ws_pcm[] = {
	//ch 0(MAIN)
	{
		.res_regs      = mn2ws_resources, 
		.n_res_regs    = ARRAY_SIZE(mn2ws_resources), 
		
		.hwd = {
			.enabled       = 1, 
			.force_aout    = mn2ws0230_pcm_aio_hwdep_force_aout, 
		}, 
		.play = {
			.enabled       = 1, 
			//0 means allocate buffer from the kernel
			.phys_addr_buf = 0x811d1000, 
			.size_buf      = MN2WS_PCM_HW_BUF_SIZE, 
			//0 means use default size
			//tentative: use 0x4000 only(restricts of HW)
			.size_buf_use  = 0x4000, 
			.setup         = mn2ws0230_pcm_aio_play_setup, 
			.mute          = mn2ws0230_pcm_aio_play_mute, 
			.start         = mn2ws0230_pcm_aio_play_start, 
			.stop          = mn2ws0230_pcm_aio_play_stop, 
			.get_hwptr     = mn2ws0230_pcm_aio_play_get_hwptr, 
			.set_devptr    = mn2ws0230_pcm_aio_play_set_devptr, 
		}, 
	}, 
	
	//ch 1(IEC out)
	{
		.res_regs      = mn2ws_resources, 
		.n_res_regs    = ARRAY_SIZE(mn2ws_resources), 
		
		.hwd = {
			.enabled       = 0, 
			.force_aout    = mn2ws0230_pcm_aio_hwdep_force_aout, 
		}, 
		.play = {
			.enabled       = 0, 
			//0 means allocate buffer from the kernel
			.phys_addr_buf = 0, 
			.size_buf      = MN2WS_PCM_HW_BUF_SIZE, 
			//0 means use default size
			.size_buf_use  = 0, 
			.setup         = mn2ws0230_pcm_aio_play_setup, 
			.mute          = mn2ws0230_pcm_aio_play_mute, 
			.start         = mn2ws0230_pcm_aio_play_start, 
			.stop          = mn2ws0230_pcm_aio_play_stop, 
			.get_hwptr     = mn2ws0230_pcm_aio_play_get_hwptr, 
			.set_devptr    = mn2ws0230_pcm_aio_play_set_devptr, 
		}, 
	}, 
	
	//ch 2(HDMI out)
	{
		.res_regs      = mn2ws_resources, 
		.n_res_regs    = ARRAY_SIZE(mn2ws_resources), 
		
		.hwd = {
			.enabled       = 0, 
			.force_aout    = mn2ws0230_pcm_aio_hwdep_force_aout, 
		}, 
		.play = {
			.enabled       = 0, 
			//0 means allocate buffer from the kernel
			.phys_addr_buf = 0, 
			.size_buf      = MN2WS_PCM_HW_BUF_SIZE, 
			//0 means use default size
			.size_buf_use  = 0, 
			.setup         = mn2ws0230_pcm_aio_play_setup, 
			.mute          = mn2ws0230_pcm_aio_play_mute, 
			.start         = mn2ws0230_pcm_aio_play_start, 
			.stop          = mn2ws0230_pcm_aio_play_stop, 
			.get_hwptr     = mn2ws0230_pcm_aio_play_get_hwptr, 
			.set_devptr    = mn2ws0230_pcm_aio_play_set_devptr, 
		}, 
	}, 
};


const char *get_pcm_dev_name(void)
{
	return MN2WS_PCM_AIO_DEV_NAME;
}

int get_pcm_dev_count(void)
{
	return MN2WS_PCM_AIO_DEV_COUNT;
}

static int mn2ws0230_pcm_aio_hwdep_force_aout(struct mn2ws_pcm_dev *dev)
{
	//do nothing
	
	return 0;
}

/**
 * ꡼׶ػ
 */
static int mn2ws0230_pcm_aio_play_setup(struct mn2ws_pcm_dev *dev)
{
	struct mem_mapping *mmp = &dev->map_regs[MN2WS_RES_PCM];
	u32 st, ed, v;
	
	st = dev->play.paddr_buf;
	ed = st + dev->play.size_alloc;
	
	//Set chunk size: 0x100(fixed by HW)
	pcm_writel(mmp, 0x100, MN2WS_APCMOUT_SAMP(dev->ch));
	
	//Dummy channel
	pcm_writel(mmp, st+0x5000, MN2WS_APCMOUT_SEND_STADDR(MN2WS_PCM_AIO_DUMMY_CH));
	pcm_writel(mmp, ed+0x5000, MN2WS_APCMOUT_SEND_ENDADDR(MN2WS_PCM_AIO_DUMMY_CH));
	
	//Dummy FS: 0x0(fixed by HW)
	pcm_writel(mmp, 0x0, MN2WS_APCMOUT_SEND_DUMMY0FS);
	
	
	//Set mode
	// 0   : Mode : Non-tunnel mode
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_MODE(dev->ch));
	
	//# ADECOUTCTL dec ch0
	pcm_writel(mmp, 0x1, MN2WS_ADECOUTCTL);

	//# DECENABLE dec ch0
	pcm_writel(mmp, 0x1, MN2WS_SYSCTRCMDA);
	
	
	v = 0;
	while ((v & (1 << 24)) == 0) {
		v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	}
	irqsh_write_intstat(v & ~(1 << 24), MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk -1-0 v=0x%08x.\n", dev->ch, v);
	v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk -1-1 v=0x%08x.\n", dev->ch, v);
	
	
	
	//Set scaler
	mn2ws0230_pcm_aio_play_mute(dev);
	
	return 0;
}

/**
 * ꡼׶ػ
 */
static int mn2ws0230_pcm_aio_play_mute(struct mn2ws_pcm_dev *dev)
{
	//Do nothing
	
	return 0;
}

/**
 * ꡼׶ػ
 */
static int mn2ws0230_pcm_aio_play_start(struct mn2ws_pcm_dev *dev)
{
	struct mem_mapping *mmp = &dev->map_regs[MN2WS_RES_PCM];
	u32 st, ed, v;
	
	DPRINTF("start the playback DMA %d.\n", dev->ch);
	
	st = dev->play.paddr_buf;
	ed = st + dev->play.size_alloc;
	
	//Set output signal
	// 0   : : 
	pcm_writel(mmp, 0x00004001, MN2WS_APCMOUT_SEND_XOCTL(dev->ch));
	
	//Set output freq
	// 0- 8: freq : 0: 48kHz
	//              1: 96kHz
	//              2: 192kHz
	//              3: 32kHz
	//              4: 44kHz
	//              5: 88kHz
	pcm_writel(mmp, 0x0, MN2WS_APCMOUT_SEND_XOFS(dev->ch));
	
	//DMA buffer address
	pcm_writel(mmp, st, MN2WS_APCMOUT_SEND_STADDR(dev->ch));
	pcm_writel(mmp, ed, MN2WS_APCMOUT_SEND_ENDADDR(dev->ch));
	
	//Configuration port for output
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_DEVSEL_CONF(dev->ch));
	
	
	//Send 1-chunk that has zero data
	pcm_writel(mmp, st + 0x400, MN2WS_APCMOUT_SEND_WRSRC(dev->ch));
	
	//Set port for output
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_DEVSEL_SEND(dev->ch));
	
	//Send chunk
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_SEMAPH(dev->ch));
	
	v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
	//DPRINTF("ch%d_p: chunk 0 semaph=0x%08x.\n", dev->ch, v);
	while (v != 0) {
		v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
	}
	//DPRINTF("ch%d_p: chunk 0 semaph=0x%08x.\n", dev->ch, v);
	//DPRINTF("ch%d_p: chunk 0 wr=0x%08x.\n", dev->ch, 
	//	pcm_readl(mmp, MN2WS_APCMOUT_SEND_WRSRC(dev->ch)));
	
	v = 0;
	while ((v & (1 << 24)) == 0) {
		v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	}
	irqsh_write_intstat(v & ~(1 << 24), MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk 0-0 v=0x%08x.\n", dev->ch, v);
	v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk 0-1 v=0x%08x.\n", dev->ch, v);
	
	
	//Send 1-chunk that has zero data
	pcm_writel(mmp, st + 0x800, MN2WS_APCMOUT_SEND_WRSRC(dev->ch));
	
	//Set port for output
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_DEVSEL_SEND(dev->ch));
	
	//Send chunk, again
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_SEMAPH(dev->ch));
	
	v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
	//DPRINTF("ch%d_p: chunk 1 semaph=0x%08x.\n", dev->ch, v);
	while (v != 0) {
		v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
	}
	//DPRINTF("ch%d_p: chunk 1 semaph=0x%08x.\n", dev->ch, v);
	//DPRINTF("ch%d_p: chunk 1 wr=0x%08x.\n", dev->ch, 
	//	pcm_readl(mmp, MN2WS_APCMOUT_SEND_WRSRC(dev->ch)));
	
	v = 0;
	while ((v & (1 << 24)) == 0) {
		v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	}
	irqsh_write_intstat(v & ~(1 << 24), MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk 1-0 v=0x%08x.\n", dev->ch, v);
	v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk 1-1 v=0x%08x.\n", dev->ch, v);
	
	
	
	//Send 1-chunk that has zero data
	pcm_writel(mmp, st + 0xc00, MN2WS_APCMOUT_SEND_WRSRC(dev->ch));
	
	//Set port for output
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_DEVSEL_SEND(dev->ch));
	
	//Send chunk, again
	pcm_writel(mmp, 0x1, MN2WS_APCMOUT_SEMAPH(dev->ch));
	
	v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
	//DPRINTF("ch%d_p: chunk 2 semaph=0x%08x.\n", dev->ch, v);
	while (v != 0) {
		v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
	}
	//DPRINTF("ch%d_p: chunk 2 semaph=0x%08x.\n", dev->ch, v);
	//DPRINTF("ch%d_p: chunk 2 wr=0x%08x.\n", dev->ch, 
	//	pcm_readl(mmp, MN2WS_APCMOUT_SEND_WRSRC(dev->ch)));
	
	v = 0;
	while ((v & (1 << 24)) == 0) {
		v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	}
	irqsh_write_intstat(v & ~(1 << 24), MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk 2-0 v=0x%08x.\n", dev->ch, v);
	v = irqsh_read_intstat(MN2WS_INTSTATAVIO6);
	//DPRINTF("ch%d_p: chunk 2-1 v=0x%08x.\n", dev->ch, v);
	
	return 0;
}

/**
 * ꡼׶ػ
 */
static int mn2ws0230_pcm_aio_play_stop(struct mn2ws_pcm_dev *dev)
{
	DPRINTF("stop the playback DMA %d.\n", dev->ch);
	
	//stop the DMA
	
	return 0;
}

/**
 * PCM  HW ɤ߼äƤХåեˤĤơ
 * ߤ HW ɤ߼֤롣
 * 
 * ХåեƬ 0 Ȥ롣
 * 
 * @param dev PCM ǥХ
 * @return HW θߤɤ߼
 */
static unsigned long mn2ws0230_pcm_aio_play_get_hwptr(struct mn2ws_pcm_dev *dev)
{
	struct mem_mapping *mmp = &dev->map_regs[MN2WS_RES_PCM];
	u32 v;
	
	v = pcm_readl(mmp, MN2WS_APCMOUT_SEND_RDSRC(dev->ch)) - 
		pcm_readl(mmp, MN2WS_APCMOUT_SEND_STADDR(dev->ch));
	
	return v;
}

/**
 * PCM  HW ɤ߼äƤХåեˤĤơ
 * ߤΥǥХν񤭹֤߰ꤹ롣
 * 
 * ХåեƬ 0 Ȥ롣
 * 
 * @param dev PCM ǥХ
 * @param ptr ǥХθߤν񤭹߰
 */
static void mn2ws0230_pcm_aio_play_set_devptr(struct mn2ws_pcm_dev *dev, unsigned long ptr)
{
	struct mem_mapping *mmp = &dev->map_regs[MN2WS_RES_PCM];
	u32 wr, wr_new, v;
	u32 st, ed;
	
	st = dev->play.paddr_buf;
	ed = st + dev->play.size_alloc;
	
	
	
	wr = pcm_readl(mmp, MN2WS_APCMOUT_SEND_WRSRC(dev->ch));
	wr_new = pcm_readl(mmp, MN2WS_APCMOUT_SEND_STADDR(dev->ch)) + ptr;
	
	if (wr != wr_new) {
		
		//Set output signal
		// 0   : : 
		pcm_writel(mmp, 0x00004001, MN2WS_APCMOUT_SEND_XOCTL(dev->ch));
		
		//Set output freq
		// 0- 8: freq : 0: 48kHz
		//              1: 96kHz
		//              2: 192kHz
		//              3: 32kHz
		//              4: 44kHz
		//              5: 88kHz
		pcm_writel(mmp, 0x0, MN2WS_APCMOUT_SEND_XOFS(dev->ch));
		
		//DMA buffer address
		pcm_writel(mmp, st, MN2WS_APCMOUT_SEND_STADDR(dev->ch));
		pcm_writel(mmp, ed, MN2WS_APCMOUT_SEND_ENDADDR(dev->ch));
		
		//Configuration port for output
		pcm_writel(mmp, 0x1, MN2WS_APCMOUT_DEVSEL_CONF(dev->ch));
		
		
		v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
		DPRINTF("ch%d_p: chunk n semaph=0x%08x.\n", dev->ch, v);
		while (v != 0) {
			v = pcm_readl(mmp, MN2WS_APCMOUT_SEMAPH(dev->ch));
		}
		DPRINTF("ch%d_p: chunk n semaph=0x%08x.\n", dev->ch, v);
		DPRINTF("ch%d_p: chunk n wr=0x%08x.\n", dev->ch, 
			wr_new);
	
		pcm_writel(mmp, wr_new, MN2WS_APCMOUT_SEND_WRSRC(dev->ch));
		
		//Set port for output
		pcm_writel(mmp, 0x1, MN2WS_APCMOUT_DEVSEL_SEND(dev->ch));
		
		//Send chunk
		pcm_writel(mmp, 0x1, MN2WS_APCMOUT_SEMAPH(dev->ch));
	}
}


static int __init mn2ws0230_pcm_aio_init_module(void)
{
	return mn2ws_pcm_init_module(mn2ws_pcm);
}

static void __exit mn2ws0230_pcm_aio_exit_module(void)
{
	mn2ws_pcm_exit_module(mn2ws_pcm);
}

module_init(mn2ws0230_pcm_aio_init_module);
module_exit(mn2ws0230_pcm_aio_exit_module);
